/* Copyright (c) 2002-2004 by Stanley M. Swanson.
   Distributed under the GNU General Public License, version 2 (GPL v2). */
/* vamp_utils.c from midivamp.c  035.26 */

# include "globals.h"

/* insertion sort: Knuth, Numerical Recipes */
void sort_i(int *v, int m)   /* single array of integers */
{ int j,k; int vt;
 for (j=1; j<m; j++) { vt = v[j];
    for (k=j-1; k>=0; k--) { if (v[k]<vt) break; v[k+1] = v[k]; }
    v[k+1] = vt; }
}

 /* sort key v, companion info u; change test to <= so that info
    with identical keys is not shuffled from original entry order 
    [ used in sort_gen_events() for midi_music_track() ] */
void sort_ii(int *v, int *u, int m) 
{ int j,k; int vt,ut;
 for (j=1; j<m; j++) { vt = v[j]; ut = u[j];
    for (k=j-1; k>=0; k--) { if (v[k]<=vt) break; 
             v[k+1] = v[k]; u[k+1] = u[k]; }
    v[k+1] = vt; u[k+1] = ut; }
}

void sort_dd(double *v, double *u, int m)
{ int j,k; double vt,ut;
 for (j=1; j<m; j++) { vt = v[j]; ut = u[j];
    for (k=j-1; k>=0; k--) { if (v[k]<vt) break;
           v[k+1] = v[k]; u[k+1] = u[k]; }
    v[k+1] = vt; u[k+1] = ut; }
}

void print_help(void)   /* print contents of help[] */
{ int i; 
  for (i=0; i<MXHELP; i++) printf("%s\n",help[i]);
      printf("\n"); }

void scan_list(int *list, char *buf)  /* recover enumerated and for lists */
{ int i,k=0,n;
  for (i=0;i<84 ;i++) { if (buf[i]==0) break; /* string terminator */
                        if (buf[i]=='=') { k = 1; buf[i] = ' '; break; } } 
  list[0] = k; 
  n = sscanf(buf," %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d",
	     list+3,list+4,list+5,list+6,list+7,list+8,list+9,list+10,
	     list+11,list+12,list+13,list+14,list+15,list+16,list+17,
	     list+18,list+19,list+20,list+21,list+22);
  list[1] = n; list[2] = 0;
  if (Q_VERBOSE>6 && V_CLASS==3) printf("scan_list type %d n==%d, last==%d\n",k,n,list[n+2]);
}


int finstring(char *ch, int max, FILE *f)
     /* reads line from FILE *f, makes 0 terminated string, returns length */
{ int i; char *got, c, z='\0';  
  if (f==NULL) { printf("bad file in finstring\n"); return EOF; } 
  got = fgets(ch,max,f); if (!got) return EOF;
  for (i=0;i<max;i++) { c = ch[i]; if (c==z) return i;
       if (c=='\n') { ch[i] = z; return i; } }
  ch[max] = z; 
  printf("WARNING: [finstring] line longer than %d \n%s\n",max,ch);
  return max; }

int next_line(void)
{  nline = finstring(line,MAXIN-4,filin); 
/*   if (Q_VERBOSE>6 && (filin!=stdin) && V_CLASS==3) printf("%s\n",line);  */
   lch = 0; cur_chr = line[lch]; if (nline<0) cur_chr = nline;
   /* 038.08  if (cur_chr==0) printf("zero length input line \n");  */
   return nline;  }  

int next_chr(void)  /* next character in line, but don't over run */
{ if (line[lch]==0 || lch >= nline) { cur_chr = 0; return 0; }  
  lch++;  cur_chr = line[lch];  return cur_chr; }

void deblank(void)
{ while (cur_chr == ' ' || cur_chr == '\t') next_chr(); }

int get_integer(int *ival) /* convert digit string into decimal integer */
{ int jval = 0, digits=0;
 while (cur_chr>='0' && cur_chr<='9')
   { jval = 10*jval + cur_chr - '0'; digits++; next_chr(); }
 if (digits) *ival = jval; return digits;
}  /* get_integer() */
 
void copy_line(void)  /* make copy of line so find_token can put '\0' to end strings */
{ int j; tok_start = token_buf; for (j=0; j<=nline; j++) token_buf[j] = line[j]; }

void lower_the_case(char *s, char *d, int n)
{ int i; char c;       /* change to lower case for key matching */
  for (i=0; i<n; i++)
    { c = s[i];  if (c>='A' && c<='Z') c += 'a' - 'A';
      d[i] = c; if (c==0) break; }
}

/* parse token_buf[] into white-space delimited strings, one at a time.  */
/* return string pointer and pointer after inserted '\0'; return NULL if no string;
   possible problem if *start is NULL  (code to catch, longjmp on error ??)  */

char *find_token(char **start) 
{ int j,k,n; char *buf=*start, *tok=NULL, c, z='\0'; 
  if (buf==NULL) { printf("NULL start in find_token()\n");    return NULL; } 
  n = (token_buf+nline) - buf;  if (n>nline) { *start = NULL; return NULL; }
  j = 0; c = *buf;
  while (c == ' ' || c == '\t')  /* remove some white space */ 
    { j++; buf++; c = *buf; if (j>n) break; }
  if (c == '\0') { *start = NULL; return NULL; }  /* end of line */
  tok = buf;
  while (c != ' ' && c != '\t' && c != '\0')  /* search for white space or end */ 
    { j++; buf++; c = *buf; if (j>n) break; }
  if (c == '\0') { *start = buf; return tok; }
  *buf = '\0'; *start = buf+1; return tok; 
}

/* match token string to list[1:n]; 0 if no match, >0 if exact, 
   <0 for "best approximate match" ("longest" partial match) */
/* 035.17 nlist<0 means search backwards (from top of list) */

int match_token(char *token, char **list, int nlist, int exact)
{ int i,j,k=0,mk=0,m,nt,ni,n, nla=nlist, jj, ambig=0; 
  char *item;
 nt = strlen(token)+1;  /* include terminating null char */
 if (nla<0) nla = -nla;
 for (jj=1; jj<nla; jj++)   /* was j<=nlist */
    {  j = jj;  if (nlist<0) j = nla - jj;  item = list[j]; 
       if (item == NULL) { if (nlist<0) continue; else break; } 
       n = ni = strlen(item)+1;
       if (n>nt) n = nt; /* search over shortest length */
       for (i=m=0; i<n; i++) if (token[i]==item[i]) m++; else break;
       if (m==n && nt==ni) return j;  /* exact match */
       if (m>mk) { mk = m; k = j; ambig = 0;}
       else if (m>0 && m==mk && exact==0) { ambig++; error_count++;
          printf(" %s ambiguous between %s and %s\n",token,list[k],item); }
    }
  if (ambig) k = 0;  /* 044.23 treat as non-match (error) Matt Kaufmann */
  return -k;  /* approximate match */
}

int save_token(char *token, char **list, int *nlist)
{ int j=0,n = *nlist;      /* transfer token to string buffer, pointer into list[] */
 if (Q_VERBOSE>6 && V_CLASS==3) printf("save_token %s   %d\n",token,n);
 list[n] = &sav_strings[n_save_chr]; *nlist = n+1;
 while (token[j] != '\0') { sav_strings[n_save_chr] = token[j];
                            j++; n_save_chr++; 
       if (j>MAXIN || n_save_chr>=MXCHAR)
       { if (flog) fprintf(flog,"*******overflow in save_token()\n"); 
         printf("******overflow in save_token()\n");  break; } } 
 sav_strings[n_save_chr] = '\0';  if (n_save_chr<MXCHAR) n_save_chr++; 
 return n; 
}

char *make_id(char *token, char *prefix, int kount)
{ if (token) return token;  /* id string furnished */
  sprintf(made_id,"%s%02d",prefix,kount);  /* make an id */
     if (Q_VERBOSE>6 && V_CLASS==3) printf("make_id ***%s***\n",made_id); 
  return made_id; }

FILE* new_file(char* name, char* how)
{ int ic=0,lc,n,k; char chr=name[0];  FILE* fi;
  while(chr==' ' && ic<MAXIN) { ic++; chr=name[ic]; } 
  if (chr=='@') { ic++; chr=name[ic]; }  /* check how[0]=='r'? */
  lc = ic; while(chr!=' ' && chr!=0) { lc++; chr=name[lc]; } 
  if (lc==ic) return NULL;
  name[lc] = 0; fi = fopen(name+ic,how);
  if (how=="r" && fi==NULL) printf("cannot open %s as %s\n",name+ic,how);
  name[lc] = chr;  return fi;
}

int stack_files(FILE* fi) 
{
       if (fi==NULL) 
	    { printf("bad file pointer (stack_file) \n"); return -1; }
      if (at_file<MXATFILE) at_file++; 
        else { if (flog) fprintf(flog,"******too many indirect files\n");  
               printf("too many indirect files\n"); return -2; over_flows++; } 
      fin[at_file] = fi; return 0;
}

int pop_file(void)
{   if (at_file>0) { fclose(fin[at_file]); at_file--; return 0; }	  
    at_file=0; return -1; /* bottom of stack */}

/* ============== from vamp_harm.c  042.20 ====================*/

int find_measures(int mel, int verbose)
{ int i,j,k,m,n,dur,per,inote,enote, pn,tn, nstart,nend,kbar=0,rep,alt;
  int ip, lp, ifrag, lfrag, kount, kf, koda=0, type, nukey, h, inkord, korknt;
 
  n = nstart = efirst[mel]; nend = elast[mel];
  dur = rep = alt = nukey = 0;   bars[0].s = nstart;
 
  if (mel<0 || mel>=chunks)
    { printf("chunk index %d out of range (find_measures)\n",mel);
      return 0; }
  type = chunk[mel].type;
    if (type != T_MEL) { printf("NOT melody chunk\n"); return 0; }
  ip = ifrag = chunk[mel].ix; lp = lfrag = chunk[mel].lx;
  kount = chunk[mel].count;  /* zero means one fragment [ip:lp-1] */
/*  if (verbose>2 && V_CLASS==5)
  printf("<<<<type %d ifrag %d lfrag %d count %d\n",type,ifrag,lfrag,kount); */
 
 /* melody fragments */
  if (kount==0) {   n = nstart = ip; nend = lp;   /*  single frag */
                    ifrag = lfrag = kf = 0;  /* dummy loop indices */ }
 
  else { if (kount<(lfrag-ifrag)) /* last fragment is coda [3 ... | */
             { lfrag = lfrag - 2; koda = lfrag; }
             /* ignore coda fragment for now... cf. gen_midi_frags()  */
      ifrag += 2;  /* 035.09 omit frag[0] (total abc) */
      kf = ifrag; n = nstart = frags[kf]; nend = frags[kf+1]; }
  while (1) {   /* 038.08  looping awkward here !!! */
  bars[kbar].s = nstart;   dur = rep = alt = nukey = 0; inkord = korknt = j = 0;   
    /*  find bar '|' delimited notes == measures */
    /* pitch in range [0,127] or E_REST==128, otherwise various marks  */
 while (n<nend) { pn = pitch[n]; tn = tics[n];
                    if (pn<0) stick[j] = 0; else stick[j] = tn; j++;
      if (pn > -1) { /* do not check for bad chord syntax */
                     if (inkord==0) dur += tn;
                     else { if (korknt==0) dur += tn; else stick[j-1] = -tn; korknt++; } }
      else if (pn==E_REP) rep = B_REP;
      else if (pn==E_REP_1 || pn==E_REP_2 || pn==E_REP_last) alt = B_END;
      else if (pn==E_KEY) nukey = B_KEY;
      else if (pn==E_START_CHORD) { inkord = 1; korknt = 0; }
      else if (pn==E_END_CHORD) { inkord = 0; korknt = 0; }
      if (pn==E_BAR && dur>0)
        { bars[kbar].e = n; bars[kbar].t = dur; bars[kbar+1].s = n + 1;
          i = alt | rep | nukey; if (dur!=ticks_in_meas) i = i | B_LEN;
if (verbose>2 && V_CLASS==5) printf(" bar %2d b %4d e %4d dur %4d marks %d\n",
            kbar,bars[kbar].s,n,dur,i);
/* 03e.07 find half measure point (may not be correct if there are chords) */
/* 030.29 will still have problems for a single note measure */
/* 030.29 old code::  k = 0; for (h=bars[kbar].s;h<n;h++) { pn = pitch[h]; tn =
tics[h];
                                  if (pn > -1) k += tn+tn;  if (k>dur) break; }
  */
           k = m = 0; for (h=bars[kbar].s; h<n && m<j; h++,m++)
             { /* pn = pitch[h]; tn = tics[h]; */  tn = stick[m];
                  if ( tn>0) k += tn+tn;  if (k>dur) break; }
           if ( k - 2*tn != dur ) {
             printf(" bad break in measure %d dur %d half %d\n",kbar+1,dur,k/2-tn);
                                    if (k -2*tn<=0 && h<n) h++;  i = i | B_SPLIT; }
   if (verbose>2 && V_CLASS==5)
           { int ii; printf("meas %2d k %4d tn %3d m %2d stick",kbar+1,k,tn,m);
             for (ii=0; ii<j; ii++) printf("%4d",stick[ii]); printf("\n"); }
 
          bars[kbar].h = h;
          bars[kbar].m = i;   kbar++;
            dur = rep = alt = nukey = 0; inkord = korknt = j = 0; }
     if (n<nend) n++; else break;
    } /* while (n<nend) */
      if (dur>0)
        {  bars[kbar].e = n; bars[kbar].t = dur;
           i = alt | rep | nukey; if (dur!=ticks_in_meas) i = i | B_LEN;
 if (verbose>3) 
   printf(" bar %2d b %4d e %4d dur %4d marks %d  end frag\n",
               kbar,bars[kbar].s,n,dur,i);
           k = m = 0; for (h=bars[kbar].s; h<n && m<j; h++,m++)
             { /* pn = pitch[h]; tn = tics[h]; */  tn = stick[m];
                  if ( tn>0) k += tn+tn;  if (k>dur) break; }
           if ( k - 2*tn != dur ) { printf(" bad break in measure %d \n",kbar);
                                    if (k -2*tn<=0 && h<n) h++;  i = i | B_SPLIT; }
           bars[kbar].h = h;
           bars[kbar].m = i; kbar++; }
   kf += 2; if (kf>=lfrag) break;  /* loop thru fragments */
   n = nstart = frags[kf]; nend = frags[kf+1];
  } /* while (1) outer fragment loop */
    return kbar;
} /* find_meas() */
 
 /* select notes from abc sequence, depending on mode (all, accent, bound) */
 /* ?? print notes if verbose and interactive ?? */
 
int select_notes(int nstart, int nend, int period, int mode, int barbreak)
{ int i,j,k,m,n,dur,per,inote,enote, pn,tn,center;
  n = nstart; dur = 0; snotes = 0; inote = enote = -1;
  per = 20000; if (period>0) per = period;
     guitar_ch_1 = guitar_ch_2 = NULL;
    /*  select notes for period or to measure bar */
    /* pitch in range [0,127], otherwise various marks  */
  while (n<nend) { pn = pitch[n];  tn = tics[n];
   if (pn > -1)
   {  dur += tn; if (pn<E_REST) { enote = n; if (inote<0) inote = n; }
      stone[snotes] = pn; stick[snotes] = tn;
      if (snotes<MXPT-1) snotes++; else over_flows++; }
 
   if (pn==E_GUITAR_CHORD)  /* remember first two guitar chords */
     { if (!guitar_ch_1) guitar_ch_1 = guitar_ch+tics[n];
       else if (!guitar_ch_2) guitar_ch_2 = guitar_ch+tics[n]; }
   if (pn==E_KEY) { set_chord_prefs(tn);
 if (/*H_verbose*/Q_VERBOSE>1 && V_CLASS==5) 
   printf("set key %d key %d mode %d\n",tn,key_major,mode_delta);
      if (snotes>0) printf("key set inside measure %d\n",smeas); }
  
      if (n<nend) n++; if (n>=nend) break;
      if (dur>=per) break;
 
   if (pn==E_BAR) { smeas++; if (dur>0 && barbreak) break; }
 /* may need better logic for lead-ins...*/
  } /* while (n<nend) */
  return n;
} /* select_notes() */
 
void select_accented(int beats) /* prune selected notes to accented ones */
{ int j,k,kt,tt,nb,nt,total=0,smallest=20000,nacc=1;
  for (j=0;j<snotes;j++)
   { k = stick[j]; total += k; if (k<smallest && k>0) smallest = k; }
  if (total != ticks_in_meas || smeas < 5)
    printf("oddball measure length %d small %d\n",total, smallest);
  if ( meter_num<meter_denom && meter_num!=3 ) nacc = 2; /* 2/4, 6/8 */
  else if (meter_num==meter_denom) nacc = 4;  /* 4/4, 2/2, C, C| */
  nt = ticks_in_meas/nacc;  /* total/nacc?? ticks between accented notes */
  nb = total/nt; kt = tt = 0;
  for (j=k=0; j<snotes; j++) { if ( tt>=kt )
         { stone[k] = stone[j]; stick[k] = stick[j]; kt += nt; k++; }
        tt += stick[j]; }
  if (smeas<5) printf("nacc %d select %d\n",nacc,k);
  snotes = k;
} /* select_accented() */
 
int abc_ticks(int ip, int lp)  /* total abc duration tics[ip] ... tics[lp-1] */
{ int i, notemark, inkord=0,  knt=0, ticks=0, dur;
 for (i=ip; i<lp; i++) { notemark = pitch[i]; dur = tics[i];
   if (notemark<0) {
     if (notemark== E_START_CHORD)
       { if (inkord) printf("extra [ in abc_ticks() at %d \n",i);
         inkord = 1; knt = 0; }
     else if (notemark== E_END_CHORD)
       { if (inkord==0 || knt==0) printf("bad ] in abc_ticks() at %d\n",i);
         inkord = knt = 0; }
   } else { if (inkord) { if (knt==0) ticks += dur; knt++;}
            else ticks += dur; }
 }
 return ticks;
} /* abc_ticks() */
 
int pedometer(int dur) /* which step (beat) in phrase */
{ /* if (dur>0 && on_beat && first_beat) printf("pedometer: begin phrase %d\n",kphrase+1); */
  int j;
  if (dur>0) { /* cumulate ticks */ tick_sum += dur; delta_ticks += dur;
    if (delta_ticks<ticks_in_beat) { off_beat = 1; on_beat = 0; }
    else { /* if (delta_ticks>ticks_in_beat) printf("excess ticks %d\n",delta_ticks); */
    /* may give problems for notes longer than one beat... */
             first_beat = last_beat = 0;
      while (delta_ticks>=ticks_in_beat) 
	{ delta_ticks -= ticks_in_beat; current_beat++;  nbeat++; }
         if (current_beat+1==beats_in_phrase) { last_beat = 1; }

      if (current_beat>=beats_in_phrase) { kphrase++;  current_beat -= beats_in_phrase;
         if (current_beat>0 || delta_ticks>0)
           printf("pedometer: end phrase %d extra beats %d ticks %d\n",
		   kphrase,current_beat,delta_ticks);
         if (current_beat == 0) first_beat = 1;

         if (kphrase>=lphrase) { /*042.28 more serious reset and check for midi generation */
	   if (delta_ticks>0) 
	     printf("end of phrase list, delta_ticks %d, nbeat %d\n",delta_ticks,nbeat);
           kphrase = iphrase;  current_beat = delta_ticks = nbeat = 0;
           down_beat = on_beat = first_beat = 1; up_beat = off_beat = last_beat = 0; }

         beats_in_phrase = phrasing[kphrase];
      } /* if (current_beat>=beats_in_phrase) change kphrase */

      if (current_beat&1) { up_beat = 1; down_beat = 0; } else { up_beat = 0; down_beat = 1; }
      if (delta_ticks==0) { on_beat = 1; off_beat = 0; }

    }  /* else if (delta_ticks>ticks_in_beat) */
  } else if (dur==0) { /* consistency check (e.g. at E_BAR) */
    if (off_beat){ printf("pedometer out of synch: beat %d delta %d\n",
                          tick_sum/ticks_in_beat+1,delta_ticks);
    printf("measure %d ",smeas+1); for (j=0;j<snotes;j++) printf(" %d",stick[j]); printf("\n");
       off_beat = 0; on_beat = 1; delta_ticks = 0; /* 044.26 limit to one measure??? */ }
     
  } else if (dur==(-1)) { /* initialize  */ tick_sum = delta_ticks = nbeat = 0;
    down_beat = on_beat = first_beat = 1; up_beat = off_beat = last_beat = 0;
    ticks_in_beat = ticks_in_meas/2;  /* presume 2 unless something in 3/4 meter */
      if (meter_num==3 && meter_denom==4) ticks_in_beat = ticks_in_meas/3;
         if (ticks_in_beat<=0) ticks_in_beat = quantum; /* if uninitialized */
    kphrase = iphrase; beats_in_phrase = phrasing[kphrase]; current_beat = 0;
      ticks_in_phrase = beats_in_phrase*ticks_in_beat;
    if (Q_VERBOSE>2 && V_CLASS==3)
    printf("pedometer() initialize: ticks_in_meas %d   ticks_in_beat %d   quantum %d\n",
           ticks_in_meas,ticks_in_beat,quantum);
  } else { /* print state  */
    printf("pedometer() state: beat %d delta %d\n",
                          tick_sum/ticks_in_beat+1,delta_ticks);
    printf("    down %d up %d on %d off %d current beat phrase %d\n",
           down_beat, up_beat, on_beat, off_beat, current_beat+1, kphrase+1);
  }
  return tick_sum/ticks_in_beat;  /* completed ordinal beats 0..63 or whatever... */
} /* pedometer() */
              
